#!/usr/bin/env node

/**
 * fpkg.js
 * 
 * Copyright (c) 2016 Mawni Maghsoudlou
 * Released under the MIT license
 */

/**
 * fpkg is a command line utility that accepts one
 * or multiple keywords, and searches the entire npm registry
 * and finds all packages relevant to given keyword(s) and, 
 * and outputs the results to the console.
 * 
 */

'use strict';

const EventEmitter = require('events');
class MyEmitter extends EventEmitter {}
const emitter = new MyEmitter();
const os = require('os');
const fs = require('fs');
const path = require('path');
const request = require('request');
const JSONStream = require('JSONStream');
const es = require('event-stream');
const columnify = require('columnify');
const col = require('chalk');
const uri = 'https://registry.npmjs.org/-/all';
const config = path.join(os.homedir(), '.fpkgrc.json');
const default_config = path.join(__dirname, '.fpkgrc.json');
const default_highcol = 'gray';

var pkgs = [];

function fetch_silent() {
  request(uri).pipe(JSONStream.parse('*')).pipe(es.mapSync((data) => {
    if (typeof data === 'object' && data.name && data.description && data['dist-tags'] && data['dist-tags'].latest) {
      emitter.emit('pkg', data);
    }
  }));
}

function parse_config() {
  try {
    return JSON.parse(fs.readFileSync(config, 'utf8'));
  } catch(err) {
    return JSON.parse(fs.readFileSync(default_config, 'utf8'));
  }
}

function highcol() {
  const conf = parse_config();
  var hcol = null;

  Object.keys(conf.highlight_color).forEach((key) => {
    if (conf.highlight_color[key]) {
      hcol = key;
    }
  });
  return hcol ? hcol : default_highcol;
}

function get_max_width() {
  var cols;
  try {
    var tty = require('tty');
    var stdout = process.stdout;
    cols = !tty.isatty(stdout.fd) ? Infinity : process.stdout.getWindowSize()[0];
    cols = (cols === 0) ? Infinity : cols;
  } catch (ex) {
    cols = Infinity; 
  }
  return cols;
}

function trim_to_max_width(str) {
  var max_width = get_max_width();
  return str.split('\n').map((line) => {  
    return line.slice(0, max_width);
  }).join('\n');
}

function get_keyword() {
  const argv = process.argv;
  var keywords = [];

  if (argv.length === 2) {
    throw new Error(`\nfpkg err: keyword cannot be null.\n`);
  } else if (argv.length === 3) {
    return argv[2].toString().toLowerCase();
  } else if (argv.length > 3) {
    argv.slice(2, argv.length).forEach((arg) => {
      keywords.push(arg.toLowerCase());
    });
    return keywords;
  }
}

function regx(keyword) {
  return new RegExp('\\b' + keyword + '\\b', 'i');
}

function highlight_search_terms(pkg, keyword) {
  var hcol = highcol();

  if (regx(keyword).test(pkg.name.toLowerCase())) {
    pkg.name = pkg.name.replace(new RegExp(regx(keyword), 'g'), col[hcol](keyword));
  }
  if (regx(keyword).test(pkg.description.toLowerCase())) {
    pkg.description = pkg.description.replace(new RegExp(regx(keyword), 'g'), col[hcol](keyword));
  }
  return pkg;
}

function colorize(pkg, keyword) {

  if (typeof keyword === 'string') {
    pkg = highlight_search_terms(pkg, keyword);
  } else {
    keyword.forEach((term) => {
      pkg = highlight_search_terms(pkg, term);
    });
  }
  return pkg;
}

function output(pkg, keyword) {

  const conf = parse_config();

  pkg.time = pkg.time.replace('T', ' ');
  pkg.time = pkg.time.replace('Z', '');
  pkg.time = pkg.time.slice(0, pkg.time.length - 4);

  if (conf.verbose) {
    pkg.description = (pkg.description.length > 70) ? `${pkg.description.slice(0, 70)}...` : pkg.description;
    console.log(`${pkg.name} -- ${pkg.description} -- ${pkg.version} -- ${pkg.author}`);
  }

  pkg = colorize(pkg, keyword); 
  pkgs.push({
    name: pkg.name,
    description: pkg.description,
    version: pkg.version,
    author: pkg.author,
    "time-modified": pkg.time
  });
}

function search_one(pkg, keyword) {
  if(regx(keyword).test(pkg.name.toLowerCase()) || 
      regx(keyword).test(pkg.description.toLowerCase()) || 
      (pkg.keywords.length > 0 && pkg.keywords.indexOf(keyword) !== -1)) {
    output(pkg, keyword);
  }
}

function search_multi(pkg, keywords) {
  var found = true;
  keywords.forEach((word) => {
    if (!regx(word).test(pkg.description.toLowerCase())) {
      found = false;
    }
  });
  if(found) {
    output(pkg, keywords);
  }
}

function fetch_and_search(keyword) {

  const conf = parse_config();
  var pkg = {};
  var timer;

  console.log(`\nfpkg: request GET ${uri}`);
  if (conf.verbose) {
    console.log('fpkg: fetching data from the npm registry..\n');
  } else {
    console.log('fpkg: fetching data from the npm registry..');
    process.stdout.write('fpkg: parsing stream of json data..');
    timer = setInterval(() => {
      process.stdout.write('.');
    }, 1000);
  }

  request(uri).pipe(JSONStream.parse('*')).pipe(es.mapSync((data) => {
    if (typeof data === 'object' && data.name && data.description && 
        data['dist-tags'] && data['dist-tags'].latest) {

      pkg.name = data.name;
      pkg.description = data.description;
      pkg.version = data['dist-tags'].latest;
      pkg.keywords = data.keywords ? data.keywords : [];
      pkg.author = (data.author && data.author.name) ? data.author.name : '';
      pkg.time = (data.time && data.time.modified) ? data.time.modified : '';

      if (typeof keyword === 'string') {
        search_one(pkg, keyword);
      } else {
        search_multi(pkg, keyword);
      }
    }
  })).on('end', () => {
    clearInterval(timer);
    console.log('\nfpkg: finished processing all packages.\n');
    if (pkgs.length > 0) {
      console.log(trim_to_max_width(columnify(pkgs, {
        truncate: true,
        columns: ['name', 'description', 'version', 'author', 'time-modified'],
        config: {
          name: { maxWidth: 40 },
          description: { maxWidth: 70 },
          version: { maxWidth: 12 },
          author: { maxWidth: 30 },
          "time-modified": { maxWidth: 19 }
        }
      }
      )));
    } else {
      console.log(`fpkg: no package found for '${keyword}'\n`);
    }
  });
}

try {
  fetch_and_search(get_keyword());
} catch(err) {
  console.error(err.message);
}

module.exports = {
  fetch: fetch_silent,
  emitter: emitter
};

